"""Descriptors encryption context"""

import binascii
import hashlib
from dataclasses import dataclass, field

from luna_cipher.cipher_factory import cipherFactory
from luna_cipher.ciphers.base_cipher import BaseCipher, InvalidKeyProvided
from luna_cipher.keys_bank.base_keys_bank import FailedToGetKey, NoKeyFound
from luna_cipher.keys_bank.vault_keys_bank import VaultKVKeysBank
from vlutils.helpers import base64ToBytes

from ..constants.derscriptor import ENCRYPTION_HASH_SIZE
from ..utils import mixins
from ..utils.log import Logger


@dataclass
class EncryptionCtx(mixins.Initializable):
    """Descriptor encryption settings"""

    # whether encryption enabled or not
    enabled: bool
    # encryption algorithm
    algorithm: str | None
    # encryption params
    params: dict | None
    # encryption key
    key: bytes | dict | None = field(init=False, default=None)
    # encryption hash (key + algorithm)
    encryptionHash: bytes | None = field(init=False, default=None)
    # cipher implementation
    cipher: BaseCipher | None = field(init=False, default=None)
    # whether contex initialized or not
    _initialized: bool = field(init=False, default=False)

    async def initialize(self):
        """Initialize"""
        if not self.enabled or self._initialized:
            return
        match self.params["source"]:
            case "vaultKV":
                key, _ = await VaultKVKeysBank(
                    url=self.params["key"]["url"], token=self.params["key"]["token"]
                ).getKey()
            case "raw":
                key = self.params["key"]
            case _:
                raise ValueError(f"Unknown encryption source {self.params['source']}")

        self.key = base64ToBytes(key)
        self.encryptionHash = hashlib.sha256(self.key + self.algorithm.encode()).digest()[:ENCRYPTION_HASH_SIZE]
        self.cipher = cipherFactory(algorithm=self.algorithm)(key=self.key)
        self._initialized = True

    async def probe(self) -> bool:
        """Check that key data is valid"""
        if not self.enabled:
            return True
        logger = Logger("Check encryption key")
        try:
            logger.info("begin")
            await self.initialize()
        except binascii.Error:
            logger.error("Failed to verify key: Bad base64 encoding")
        except InvalidKeyProvided as err:
            logger.error(f"Failed to verify key: {str(err)}")
        except FailedToGetKey as err:
            logger.error(f"Request to remote source failed: {str(err)}")
        except NoKeyFound as err:
            logger.error(f"Failed to get key from remote source: {str(err)}")
        else:
            logger.info("success")
            return True
        return False
