"""Redis storage context"""

from typing import Awaitable, List, Protocol, Tuple

from ..utils.check_connection import checkConnectionToRedis, checkConnectionToSentinel
from ..utils.healthcheck import checkRedis
from ..utils.mixins import Initializable
from .adaptor import RedisAdaptor


class RedisStorageSetting(Protocol):
    """Redis storage settings protocol"""

    class Sentinel:
        """Redis sentinel settings"""

        masterName: str
        sentinels: List[Tuple[str, int]]
        user: str
        password: str

    host: str
    port: int
    user: str
    password: str
    dbNumber: int
    sentinel: Sentinel


class RedisContext(Initializable):
    """Redis db context"""

    def __init__(self, redisSettings: RedisStorageSetting):
        self._redisSettings = redisSettings
        self._adaptor = RedisAdaptor(
            host=redisSettings.host,
            port=redisSettings.port,
            username=redisSettings.user,
            password=redisSettings.password,
            dbNumber=redisSettings.dbNumber,
            sentinels=redisSettings.sentinel.sentinels,
            masterName=redisSettings.sentinel.masterName,
            sentinelUsername=redisSettings.sentinel.user,
            sentinelPassword=redisSettings.sentinel.password,
        )

    async def probe(self):
        """
        Probe redis config
        """
        dbSettings = self._redisSettings

        if not dbSettings.sentinel.sentinels:
            return await checkConnectionToRedis(
                redisHost=dbSettings.host,
                redisPort=dbSettings.port,
                redisUser=dbSettings.user,
                redisPassword=dbSettings.password,
                asyncCheck=True,
            )

        return await checkConnectionToSentinel(
            sentinels=dbSettings.sentinel.sentinels,
            masterName=dbSettings.sentinel.masterName,
            sentinelUser=dbSettings.sentinel.user,
            sentinelPassword=dbSettings.sentinel.password,
            redisUser=dbSettings.user,
            redisPassword=dbSettings.password,
            asyncCheck=True,
        )

    def getRuntimeChecks(self, _) -> list[tuple[str, Awaitable]]:
        """Get redis storage runtime checks"""
        return [("redis_storage", checkRedis(self.client))]

    async def close(self):
        """Close redis adaptor"""
        await self._adaptor.close()

    @property
    def client(self):
        """Get adaptor client."""
        return self._adaptor.client
