"""
LunaApplication is an aggregate of objects required to start and configure service.
It wraps sanic application, effectively separating namespaces.

Attributes that are meant to be initialized and accessed during runtime should be
set in context class.
"""

import signal
from pathlib import Path
from typing import Type

import log
from configs_core.base_service_setting import BaseServiceSettingsClass
from luna_plugins.base.manager import PluginManager
from sanic import Sanic as BaseSanic
from sanic.handlers import ErrorHandler
from sanic.http.tls.context import create_context
from sanic.mixins.listeners import ListenerEvent
from sanic.views import HTTPMethodView
from sanic.worker.loader import AppLoader

from ..constants.http_consts import (
    STANDARD_KEEP_ALIVE_TIMEOUT,
    STANDARD_REQUEST_MAX_SIZE,
    STANDARD_REQUEST_TIMEOUT,
    STANDARD_RESPONSE_TIMEOUT,
)
from ..constants.paths import APP_ROOT
from ..constants.third_party_loggers import THIRD_PARTY_LOGGERS
from ..errors.errors import Error
from ..errors.exception import VLException
from ..utils import mixins
from ..utils.log import logger
from .cmd_parser import ApplicationCMDParser
from .context import LunaAppCtx
from .error_handlers import serverErrorHandler
from .gc_tuning import gcTuning
from .metrics_handler import MetricsHandler
from .middlewares import (
    compressResponseMiddleware,
    deCompressResponseMiddleware,
    endRequestMiddleware,
    logErrorMiddleware,
    restMiddleware,
    sendToMonitoringMiddleware,
)
from .request import VLRequest
from .standard_handlers.plugins_handler import PluginsHandler


class LunaApp[TConfig: BaseServiceSettingsClass, TCtx: LunaAppCtx]:
    """Luna service web application"""

    ctxCls: Type[TCtx]
    configCls: TConfig
    defaultPort: int
    apiVersion: int
    requestCls = VLRequest

    def __init__(
        self,
        thirdPartyLoggerWhiteList: list[str] | None = None,
        serviceConfig: TConfig | None = None,
        defaultPort: int | None = None,
    ):
        if not thirdPartyLoggerWhiteList:
            thirdPartyLoggerWhiteList = THIRD_PARTY_LOGGERS
        else:
            thirdPartyLoggerWhiteList.extend(THIRD_PARTY_LOGGERS)

        if defaultPort is not None:
            self.defaultPort = defaultPort

        self.cmdArgs = cmdArgs = self._getCMDArgParser().parse()

        if not serviceConfig:
            serviceConfig = self.loadConfig()
        self.serviceConfig = serviceConfig
        self.configureLogger(cmdArgs.logSuffix, thirdPartyLoggerWhiteList)

        # compat
        self.name = serviceConfig.serviceName

        sanic = self.sanic = Sanic(
            name=self.name,
            strict_slashes=False,
            request_class=self.requestCls,
            configure_logging=False,
            error_handler=LunaErrorHandler(),
        )
        self.setUpSanic()

        self.standardRoutes = {}

        self.setStandardRoutes()

        pluginManager = PluginManager(self, self.pluginDirectory)

        async def _initializeCtx(*_):
            """
            Initialize components. Order matters.
            Pluginmanager always in the end.
            """

            sanic.ctx = ctx = self.ctxCls(
                serviceConfig,
                pluginManager,
                apiVersion=self.apiVersion,
            )  # pylint: disable-msg=attribute-defined-outside-init
            ctx.cmdArgs = cmdArgs

            try:
                await ctx.initialize()
            except mixins.InitializationError:
                self.sanic.stop()

            # plugins have inclusive access to ctx, initializing them after others
            await pluginManager.initialize(serviceConfig.activePlugins)
            await pluginManager.start()

        sanic.register_listener(_initializeCtx, ListenerEvent.BEFORE_SERVER_START.name)

        async def _shutdown(*_):
            """
            Shutdown components from leafs to roots.
            """
            await pluginManager.shutdown()
            await sanic.ctx.close()

        sanic.register_listener(_shutdown, ListenerEvent.AFTER_SERVER_STOP.name)

        async def _printServiceStart(app, loop):  # pylint: disable-msg=W0613
            """
            Print a message about that service is started
            """
            host, port = sanic.state.host, sanic.state.port
            host = host if host != "0.0.0.0" else "127.0.0.1"
            protocol = "https" if self.cmdArgs.tlsCert and self.cmdArgs.tlsKey else "http"
            print(f"======== Running on {protocol}://{host}:{port} ========")
            logger.info(f"Service is running on {protocol}://{host}:{port}")

        sanic.register_listener(_printServiceStart, ListenerEvent.AFTER_SERVER_START.name)

        gcTuning()

    def setUpSanic(self):
        """Set up sanic"""
        sanic = self.sanic
        sanic.register_middleware(deCompressResponseMiddleware, "request")
        sanic.register_middleware(endRequestMiddleware, "response")
        sanic.register_middleware(restMiddleware, "response")
        sanic.register_middleware(compressResponseMiddleware, "response")
        sanic.register_middleware(sendToMonitoringMiddleware, "response")
        sanic.register_middleware(logErrorMiddleware, "response")

        # see https://sanic.readthedocs.io/en/latest/sanic/config.html#builtin-configuration-values for details
        if httpSettings := getattr(self.serviceConfig, "httpSettings", None):
            requestTimeout = httpSettings.requestTimeout
            responseTimeout = httpSettings.responseTimeout
            requestMaxSize = httpSettings.requestMaxSize
            keepAliveTimeout = httpSettings.keepAliveTimeout
        else:
            requestTimeout = STANDARD_REQUEST_TIMEOUT
            responseTimeout = STANDARD_RESPONSE_TIMEOUT
            requestMaxSize = STANDARD_REQUEST_MAX_SIZE
            keepAliveTimeout = STANDARD_KEEP_ALIVE_TIMEOUT
        sanic.config["REQUEST_TIMEOUT"] = requestTimeout
        sanic.config["RESPONSE_TIMEOUT"] = responseTimeout
        sanic.config["REQUEST_MAX_SIZE"] = requestMaxSize
        sanic.config["KEEP_ALIVE_TIMEOUT"] = keepAliveTimeout
        sanic.config["WEBSOCKET_PING_INTERVAL"] = None
        sanic.config["ACCESS_LOG"] = False

    @property
    def ctx(self) -> TCtx:
        """Get current service context"""
        return self.sanic.ctx

    def _getCMDArgParser(self) -> ApplicationCMDParser:
        """
        Get command line arguments parser
        Returns:
            parser with standard  options
        """
        argParser = ApplicationCMDParser(self.defaultPort, tagableSettings=self.configCls.getServiceSettingNames())
        return argParser

    def loadConfig(self) -> TConfig:
        """Fetch and apply configuration."""
        serviceConfig = self.configCls()
        serviceConfig.load(
            configFile=self.cmdArgs.config,
            configuratorAddress=self.cmdArgs.configuratorAddress,
            tags=self.cmdArgs.tags,
            asyncLoad=False,
        )
        return serviceConfig

    def configureLogger(self, logSuffix: str, thirdPartyLoggerWhiteList: list[str]):
        """Given cmdArgs and serviceConfig, perform logging configuration."""
        config = self.serviceConfig
        log.initiate(
            appName=config.serviceName,
            logLevel=config.logger.logLevel,
            logTime=config.logger.logTime,
            folderForLog=config.logger.folderWithLogs,
            maxSize=config.logger.maxLogFileSize,
            logNameSuffix=logSuffix,
            logToStdOut=config.logger.logToStdOut,
            logToFile=config.logger.logToFile,
            multilineStackTrace=config.logger.multilineStackTrace,
            thirdPartyLoggerWhiteList=thirdPartyLoggerWhiteList,
            formatter=config.logger.format,
        )

    def setStandardRoutes(self):
        """Set standard application routes"""
        self.standardRoutes = {rf"/{self.apiVersion}/plugins": PluginsHandler, r"/metrics": MetricsHandler}

    def addStandardRoutes(self):
        """Add standard application routes"""
        self.addRoutes(list(self.standardRoutes.items()))

    def addRoutes(self, routes: list[tuple[str, Type[HTTPMethodView]]]):
        """
        Add routes.

        This method accepts type or view function as view.

        Args:
            routes: list of routes (<resource, HTTPMethodView>)
        """
        for resource, view in routes:
            if hasattr(view, "view_class"):
                self.sanic.add_route(view, resource)
            else:
                self.sanic.add_route(view.as_view(), resource)

    def run(self, host="0.0.0.0", **kwargs):
        """Run sanic application."""
        self.addStandardRoutes()

        app = AppLoader(factory=lambda: self.sanic).load()

        if self.cmdArgs.tlsCert and self.cmdArgs.tlsKey:
            kwargs["ssl"] = create_context(
                certfile=self.cmdArgs.tlsCert, keyfile=self.cmdArgs.tlsKey, password=self.cmdArgs.tlsKeyPass
            )

        sock = kwargs.pop("sock", None)
        if sock:
            _, port = sock.getsockname()
            app.prepare(host=host, port=port, **kwargs)
            # if sock is specified sanic wont bind a new one
            # otherwise app will get stuck and tests wont pass
            app.state.server_info[0].settings["sock"] = sock
        else:
            port = kwargs.pop("port", None) or self.cmdArgs.port
            app.prepare(host=host, port=port, **kwargs)
        Sanic.serve_single(primary=app)

    def updateStandardRoutes(self, routes: dict[str, Type[HTTPMethodView]]):
        """
        Update LunaApp standard routes

        Args:
            routes: routes to update
        """
        self.standardRoutes.update(routes)

    @property
    def pluginDirectory(self) -> Path:
        """Get plugin directory"""
        return APP_ROOT / "plugins"


class LunaErrorHandler(ErrorHandler):
    """Common error handler"""

    def default(self, request: VLRequest, exception):
        """Default error handler."""
        return serverErrorHandler(request, exception)


class Sanic(BaseSanic):  # pylint: disable-msg=too-many-ancestors
    """Sanic customization."""

    _allow_route_overwrite = True

    async def _static_request_handler(self, *args, **kwargs):
        """
        Overrides sanic static processing because it returns None on `IsADirectory` exception that leads to 500 error
        """
        result = await super()._static_request_handler(*args, **kwargs)  # pylint: disable-msg=missing-kwoa
        if result is None:
            raise VLException(Error.PageNotFoundError, 404, False)
        return result

    async def _websocket_handler(
        self, handler: "WSBaseHandler", request: VLRequest, *args, subprotocols=None, **kwargs
    ):
        """
        Websocket handle. Open ws session with validation and preparation.
        """
        wsSession = getattr(handler, "wsSession", None)  # hack for data validation before handshake
        kwargs2Send = {} if getattr(handler, "ignorePathParams", False) is True else kwargs
        if wsSession:
            handler = handler(request, wsSession)
            async with handler.wsSession(request):
                await super()._websocket_handler(
                    lambda r, w: handler.get(w), request, *args, subprotocols=subprotocols, **kwargs2Send
                )
        else:
            await super()._websocket_handler(
                lambda r, w: handler.get(w), request, *args, subprotocols=subprotocols, **kwargs2Send
            )

    def stop(self, *args, **kwargs):
        """
        Stops sanic application, taking care of not shutting down twice.
        """
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        super().stop(*args, **kwargs)
