"""
Configs Handler.

Module realizes configs handler.
"""

from sanic.response import HTTPResponse

from ..errors.errors import Error
from ..errors.exception import VLException
from ..web.handlers import BaseHandler
from .application import LunaApp


class ConfigsHandler(BaseHandler):
    """
    Handler for getting actual service configuration
    """

    @property
    def config(self):
        """
        Get running application config
        Returns:
            running application config
        """
        return self.request.config

    async def get(self) -> HTTPResponse:  # pylint: disable-msg=W0221
        """
        Get service configuration, see `get_config`_.

        .. _get_config:
            _static/api.html#operation/getConfig

        Resource is reached by address '/config'

        Raises:
            VLException(Error.UnsupportedMediaType, 415, isCriticalError=False), if "Accept" header is incorrect
        """
        acceptHeaders = [row.split(";")[0] for row in self.request.headers.get("Accept", "application/json").split(",")]
        acceptJson = "application/json" in acceptHeaders
        acceptText = "text/plain" in acceptHeaders
        acceptAny = "*/*" in acceptHeaders

        if acceptJson or acceptAny:
            config: dict = self.config.getConfigAsFlattenDict()
            return self.success(outputJson=config, contentType="application/json")

        if acceptText:
            config: str = self.config.getConfigAsText()
            return self.success(body=config, contentType="text/plain")

        raise VLException(Error.UnsupportedMediaType, 415, isCriticalError=False)


def addConfigHandlerRoutes(app: LunaApp) -> None:
    """
    Add standard configs handler to app
    Args:
        app: application
    """
    apiVersion = app.apiVersion
    app.sanic.add_route(ConfigsHandler.as_view(), rf"/{apiVersion}/config", strict_slashes=True)
